/*
 * PS/2 to X68000 keyboard converter.
 *
 * Copyright (c) 2012, Lyderic Maillet <lydux86@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The x68000 keyboard port is directly connected to MFP usart configured
 * by IPL to 2400bps, 8 bits, 1 stop bit and no parity.
 * This code only perform scancodes translation from generic PS/2 keyboard
 * to X68k and deal with MFP receiver line. It doesn't perform commands
 * conversion from X68k to PS/2 (like leds handling or setting keys
 * repeat/delay time.
 *
 * The following code is designed for an ATMEGA32 using avr-gcc and avr-libc.
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>

#ifndef F_CPU
#define F_CPU 	11059200UL
#endif
#define BAUD    2400

#include <util/setbaud.h>
#include <util/delay.h>

#define X68K_DDR	DDRA
#define X68K_PIN	PINA
#define X68K_READY	PA0

#define PS2_DDR		DDRD
#define PS2_PIN		PIND
#define PS2_DAT		PD3

#include "keymap.h"

void usart_init (void)
{
	UBRRH = UBRRH_VALUE;
	UBRRL = UBRRL_VALUE;

	UCSRA = 0;
	UCSRB = _BV (TXEN) | _BV (RXEN);
	UCSRC = _BV (URSEL) | _BV (UCSZ1) | _BV (UCSZ0);
}

void port_init (void)
{
	/* PS/2 DAT as input */
	PS2_DDR &= ~_BV (PS2_DAT);

	GICR |= _BV (INT0);
	MCUCR |= _BV (ISC01);

	/* X68K READY signal as input */
	X68K_DDR &= ~_BV (X68K_READY);
}

void usart_tx (uint8_t v)
{
	loop_until_bit_is_set (UCSRA, UDRE);
	UDR = v;
}

void send_key (uint8_t v)
{
	/* Disable interrupt, avoid receiving a new keystroke without
	 * sending the last one */
	cli ();

	/* Wait for MFP receiver to flush its content */
	while (!(X68K_PIN & _BV (X68K_READY)));

	usart_tx (v);

	/* Enable interrupt */
	sei ();
}

#define PS2_BREAK_KEY	0xF0
#define PS2_KEYTYPE_MODIFIER	0xE0

enum key_action {
	KEY_DOWN,
	KEY_UP
};

struct key_state {
	uint8_t		scancode;
	enum key_action	action;
	const uint8_t *	lut;
};

/* Convert to X68K keystroke, and send the result
 * through UART. */
void tox68k (struct key_state *key)
{
	uint8_t code;

	/* Perform keystroke translation */
	code = key->lut[key->scancode];

	/* No translation for this key. Don't care... */
	if (code == 0x00)
		return;

	/* Key released */
	if (key->action == KEY_UP)
		code |= 0x80;
	
	send_key (code);
}

/* Process for the decoding of a PS/2 keystroke.
 * Handle standard and extended scancodes set. 
 *
 * TODO : handle funky keystrokes like PAUSE or 
 *        PRINT SCREEN. */
void ps2_process_keystroke (uint8_t scancode)
{
	static struct key_state key;

	if (scancode == PS2_BREAK_KEY)
	{
		key.action = KEY_UP;
	}
	else if (scancode == PS2_KEYTYPE_MODIFIER)
	{
		key.lut = keymap_mod_tab;
	}
	else
	{
		if (scancode < 0x90)
		{
			key.scancode = scancode;
			tox68k (&key);
		}

		key.lut = keymap_at_tab;
		key.action = KEY_DOWN;
		key.scancode = 0;
	}
}

/* This is the PS/2 clock interception routine.
 * It should be called whenever clock fall to a logic 0. */
ISR (INT0_vect)
{
	static uint8_t _data = 0, _b, _par = 0;
	static int _seq = -1;

	_seq++;

	/* Get PS/2 data bit */
	_b = (PS2_PIN & _BV (PS2_DAT)) >> PS2_DAT;

	/* Compute parity bit */
	if ((_seq > 0) && (_seq < 10))
		_par ^= _b;

	if ((_seq == 0) && (!_b))
		/* Start bit, always 0 */
		_data = 0;
	else if ((_seq > 0) && (_seq < 9))
		/* Data bits */
		_data |= (_b << (_seq-1));
	else if ((_seq == 9) && (_par))
		/* Parity ok (odd) */
		_par = 0;
	else if ((_seq == 10) && (_b))
	{
		/* Stop bit, always 1 */
		ps2_process_keystroke (_data);
		_seq = -1;
	}
	else
	{
		/* Something's wrong ! We should never get here. Just put
		 * this code in case of...
		 * Forget all and wait for 10ms. Should be long enough to
		 * give up the few following bits and fast enough before
		 * the next incoming keystroke. */
		_seq = -1;
		_delay_ms (10);
	}

}

int main (void)
{
	/* Init plaform */
	usart_init ();
	port_init ();

	/* Enable interrupt, start waiting for PS/2 clock. */
	sei ();

	/* Wait forever */
	while(1);

	return 0;
}

